from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

from config_keeper.console import DefaultHighlighter


def spinner() -> Progress:
    return Progress(
        SpinnerColumn('dots', style='bold yellow'),
        TimeElapsedColumn(),
        TextColumn('{task.description}', highlighter=DefaultHighlighter()),
        transient=True,
    )



import typer

from config_keeper import settings
from config_keeper.console_helpers import print_error, print_tip


class ConfigKeeperError(Exception):
    pass


class InternalError(ConfigKeeperError):
    pass


class PublicError(typer.Exit, ConfigKeeperError):
    exit_code = 255  # default

    def __init__(self, msg: str | None = None, *, tip: str | None = None):
        if msg:
            print_error(msg)
        if tip:
            print_tip(tip)
        super(ConfigKeeperError, self).__init__(msg)


class InvalidConfigError(PublicError):
    exit_code = 201


class ProjectDoesNotExistError(PublicError):
    exit_code = 203

    def __init__(self, project: str, *, tip: str | None = None):
        self.project = project
        super().__init__(f'project "{project}" does not exist.', tip=tip)


class AtLeastOneOptionMustBeProvidedError(PublicError):
    exit_code = 204

    def __init__(self, *, tip: str | None = None):
        super().__init__('at least one option must be provided.', tip=tip)


class InvalidArgumentFormatError(PublicError):
    exit_code = 205


class DuplicatePathNameError(PublicError):
    exit_code = 207

    def __init__(self, name: str, *, tip: str | None = None):
        super().__init__(
            f'path name "{name}" is repeated multiple times.',
            tip=tip,
        )


class PathNameAlreadyInProjectError(PublicError):
    exit_code = 208

    def __init__(self, name: str, project: str, *, tip: str | None = None):
        super().__init__(
            f'path name "{name}" already in "{project}".',
            tip=tip,
        )


class PathNameDoesNotExistError(PublicError):
    exit_code = 209

    def __init__(self, name: str, project: str, *, tip: str | None = None):
        super().__init__(
            f'project "{project}" does not have path named "{name}".',
            tip=tip,
        )


class SyncError(PublicError):
    exit_code = 220

    def __init__(self, errors_map: dict[str, str]):
        msg = 'operation failed for following projects:\n\n'

        for project, error in errors_map.items():
            msg += f'{project}\n'
            msg += '    ' + error.replace('\n', '\n    ') + '\n'

        super().__init__(msg)


class ExecutableNotFoundError(PublicError):
    exit_code = 254

    def __init__(self, executable: str):
        super().__init__(
            f'executable "{executable}" is not found in your system. '
            f'It is required for {settings.EXECUTABLE_NAME} to work correctly.',
        )



import subprocess  # noqa: I001
import typing as t
import importlib.metadata

import typer

from config_keeper.commands.config import cli as config_cli
from config_keeper.commands.paths import cli as paths_cli
from config_keeper.commands.project import cli as project_cli
from config_keeper import config, console, settings
from config_keeper import exceptions as exc
from config_keeper.sync_handler import SyncHandler
from config_keeper.validation import ProjectValidator
from config_keeper.progress import spinner

if t.TYPE_CHECKING:
    from rich.progress import TaskID

cli = typer.Typer(
    name=settings.EXECUTABLE_NAME,
    help='CLI tool for keeping your personal config files in a repository',
)

cli.add_typer(project_cli, name='project', help='Manage projects.')
cli.add_typer(config_cli, name='config', help='Manage config of this tool.')
cli.add_typer(paths_cli, name='paths', help='Manage project paths.')


@cli.callback(invoke_without_command=True)
def print_version(
    version: t.Annotated[
        bool,
        typer.Option(
            '--version',
            help='Show current version and exit.',
            is_eager=True,
        ),
    ] = False,
):
    if version:
        console.print(
            importlib.metadata.version('config-keeper2'),
            highlight=False,
        )
        raise typer.Exit


@cli.command()
def push(
    projects: t.List[str],  # noqa: UP006
    ask: t.Annotated[bool, typer.Option()] = True,
):
    """
    Push files or directories of projects to their repositories. This operation
    is NOT atomic (i.e. failing operation for some project does not prevent
    other projects to be processed).
    """

    conf = config.load()
    validator = ProjectValidator(conf, path_existence='error')

    _validate_projects(projects, validator)

    if ask:
        console.print('Going to push into following branches:')

        for project in projects:
            console.print(
                f'- "{conf["projects"][project]["branch"]}" at '
                f'{conf["projects"][project]["repository"]} '
                f'(from "{project}")',
            )

        if not typer.confirm('Proceed?', default=True):
            raise typer.Exit

    _operate('push', projects, conf)


@cli.command()
def pull(
    projects: t.List[str],  # noqa: UP006
    ask: t.Annotated[bool, typer.Option()] = True,
):
    """
    Pull all files and directories of projects from their repositories and move
    them to projects' paths with complete overwrite of original files. This
    operation is NOT atomic (i.e. failing operation for some project does not
    prevent other projects to be processed).
    """

    conf = config.load()
    validator = ProjectValidator(conf, path_existence='skip')

    _validate_projects(projects, validator)

    if ask:
        console.print('Following paths will most likely be replaced:')

        for project in projects:
            for path in conf['projects'][project]['paths'].values():
                console.print(f'- {path} (from "{project}")')

        if not typer.confirm('Proceed?', default=True):
            raise typer.Exit

    _operate('pull', projects, conf)


def _validate_projects(
    projects: list[str],
    validator: ProjectValidator,
):
    with spinner() as s:
        prev_task: TaskID | None = None
        for project in projects:
            if prev_task is not None:
                s.stop_task(prev_task)
            prev_task = s.add_task(
                f'Validating project "{project}"...',
                total=None,
            )
            validator.validate(project)

    validator.print_errors()
    if not validator.is_valid:
        raise exc.InvalidConfigError


def _operate(
    operation: t.Literal['push', 'pull'],
    projects: list[str],
    conf: config.TConfig,
):
    errors: dict[str, str] = {}

    with spinner() as s:
        prev_task: TaskID | None = None
        for project in projects:
            if prev_task is not None:
                s.stop_task(prev_task)
            prev_task = s.add_task(
                f'Processing project "{project}"...',
                total=None,
            )
            handler = SyncHandler(project, conf)
            try:
                getattr(handler, operation)()
            except subprocess.CalledProcessError as e:
                errors[project] = e.stdout + e.stderr

    if errors:
        raise exc.SyncError(errors)

    console.print('Operation successfully completed.')



verbose = 'Show additional information.'
project = """
    The name of project.
"""
repository = """
    Repository which is used to store your files and directories.
"""
branch = """
    Branch of the repository used to push and pull from.
"""
check = """
    Whether check if repository exist.
"""



from config_keeper import console


def print_project_saved(project: str):
    console.print(f'Project "{project}" saved.')


def print_warning(msg: str):
    console.print('Warning:', end=' ', style='yellow')
    console.print(msg)


def print_error(msg: str):
    console.print('Error:', end=' ', style='red')
    console.print(msg)


def print_critical(msg: str):
    console.print('Critical:', end=' ', style='violet')
    console.print(msg)


def print_tip(msg: str):
    console.print('Tip:', end=' ', style='yellow')
    console.print(msg)



import subprocess
import typing as t

import typer
import yaml

from config_keeper import config, console, settings
from config_keeper import exceptions as exc
from config_keeper.commands import helps
from config_keeper.console_helpers import print_error, print_project_saved
from config_keeper.validation import check_if_project_exists, ping_remote

cli = typer.Typer()


@cli.command()
def create(
    project: t.Annotated[str, typer.Argument(help=helps.project)],
    repository: t.Annotated[
        str,
        typer.Option(prompt=True, help=helps.repository),
    ],
    branch: t.Annotated[
        str,
        typer.Option(prompt=True, help=helps.branch),
    ] = 'main',
    check: t.Annotated[bool, typer.Option(help=helps.check)] = True,
):
    """
    Create a new project.
    """

    if check:
        _check_remote(repository)

    conf = config.load()

    if project in conf['projects'] and not typer.confirm(
        f'Project "{project}" already exists. '
        'Would you like to overwrite it?',
    ):
        raise typer.Exit

    conf['projects'][project] = {
        'repository': repository,
        'branch': branch,
        'paths': {},
    }
    config.save(conf)
    print_project_saved(project)


@cli.command()
def update(
    project: t.Annotated[str, typer.Argument(help=helps.project)],
    repository: t.Annotated[
        t.Optional[str],  # noqa: UP007
        typer.Option(help=helps.repository),
    ] = None,
    branch: t.Annotated[
        t.Optional[str],  # noqa: UP007
        typer.Option(help=helps.branch),
    ] = None,
    check: t.Annotated[bool, typer.Option(help=helps.check)] = True,
):
    """
    Update project.
    """

    conf = config.load()
    check_if_project_exists(project, conf)

    updated = False
    data = conf['projects'][project]

    if repository:
        if check:
            _check_remote(repository)
        data['repository'] = repository
        updated = True

    if branch:
        data['branch'] = branch
        updated = True

    if not updated:
        raise exc.AtLeastOneOptionMustBeProvidedError

    config.save(conf)
    print_project_saved(project)


@cli.command()
def delete(
    project: t.Annotated[str, typer.Argument(help=helps.project)],
    confirm: t.Annotated[
        bool,
        typer.Option(help='Whether confirm before deleting.'),
    ] = True,
):
    """
    Delete project.
    """

    conf = config.load()
    check_if_project_exists(project, conf)

    if confirm:
        console.print('You are about delete the following project:\n')
        console.print(yaml.dump({project: conf['projects'][project]}))
        if not typer.confirm('Confirm?'):
            raise typer.Exit

    del conf['projects'][project]
    config.save(conf)
    console.print(f'Project "{project}" deleted.')


@cli.command()
def show(project: t.Annotated[str, typer.Argument(help=helps.project)]):
    """
    Show project config.
    """

    conf = config.load()
    check_if_project_exists(project, conf)

    info = {project: conf['projects'][project]}
    console.print(yaml.dump(info))


@cli.command(name='list')
def list_(
    verbose: t.Annotated[
        bool,
        typer.Option('--verbose', '-v', help=helps.verbose),
    ] = False,
):
    """
    List all projects.
    """

    conf = config.load()

    if not conf['projects']:
        console.print('You have no projects. Create a new one using')
        console.print(f'> {settings.EXECUTABLE_NAME} project create')

    if verbose:
        console.print(yaml.dump(conf['projects']))
    else:
        for project in conf['projects']:
            console.print(project)


def _check_remote(repository: str):
    console.print(f'Checking {repository}...', end=' ')
    try:
        ping_remote(repository, log=True)
        console.print('OK', style='green')
    except subprocess.CalledProcessError as exc:
        print_error(f'\n\n{exc.stderr}')
        if not typer.confirm('Do you want to continue?'):
            raise typer.Exit from exc



import datetime
import shutil
import subprocess
import tempfile
from pathlib import Path

from config_keeper import config
from config_keeper import exceptions as exc


def delete_dir(directory: str | Path):
    """
    Deletes the whole directory if it exists.
    """
    assert Path(directory).is_dir()
    shutil.rmtree(str(directory), ignore_errors=True)


def run_cmd(cmd: list[str]) -> subprocess.CompletedProcess:
    executable = shutil.which(cmd[0])
    if not executable:
        raise exc.ExecutableNotFoundError(cmd[0])
    cmd = [executable] + cmd[1:]
    return subprocess.run(cmd, check=True, capture_output=True, text=True)


def clear_working_tree(path: Path | str):
    """
    Removes all files and directories in specified path except .git directory.
    """
    for entity in Path(path).iterdir():
        if entity.stem == '.git':
            continue
        if entity.is_dir():
            shutil.rmtree(entity)
        elif entity.is_file():
            entity.unlink()


def remote_branch_exists(
    repo: Path | str,
    branch: str,
    remote_name: str = 'origin',
) -> bool:
    full_name = f'refs/heads/{branch}'
    result = run_cmd([
        'git', '-C', str(repo), 'ls-remote', '--heads', remote_name,
        full_name,
    ])
    return full_name in result.stdout


class SyncHandler:
    def __init__(self, project: str, conf: config.TConfig):
        self.project = project
        self.conf = conf

    def push(self):
        branch = self.conf['projects'][self.project]['branch']
        repository = self.conf['projects'][self.project]['repository']

        temp_dir = tempfile.mkdtemp()

        run_cmd(['git', 'init', temp_dir])
        run_cmd(['git', '-C', temp_dir, 'remote', 'add', 'origin', repository])

        if remote_branch_exists(temp_dir, branch):
            run_cmd(['git', '-C', temp_dir, 'fetch', 'origin'])
            run_cmd(['git', '-C', temp_dir, 'checkout', branch])
            clear_working_tree(temp_dir)
        else:
            run_cmd(['git', '-C', temp_dir, 'checkout', '-b', branch])

        self._fetch_files(temp_dir)

        run_cmd(['git', '-C', temp_dir, 'add', '.'])
        run_cmd([
            'git', '-C', temp_dir, 'commit', '-m', self._get_commit_message(),
        ])
        run_cmd([
            'git', '-C', temp_dir, 'push', '--set-upstream', 'origin', branch,
        ])

        delete_dir(temp_dir)

    def pull(self):
        pull_dir = tempfile.mkdtemp()
        branch = self.conf['projects'][self.project]['branch']
        repository = self.conf['projects'][self.project]['repository']

        run_cmd(['git', 'init', pull_dir])
        run_cmd([
            'git', '-C', pull_dir, 'pull', repository, branch,
        ])

        self._put_in_places(pull_dir)
        delete_dir(pull_dir)

    def _fetch_files(self, directory: Path | str):
        directory = Path(directory)

        paths = self.conf['projects'][self.project]['paths']
        for path_name, str_path in paths.items():
            path = Path(str_path)
            dest = str(directory / path_name)
            copy_args = (str_path, dest)
            if path.is_file():
                shutil.copy2(*copy_args)
            else:
                shutil.copytree(*copy_args)

    def _put_in_places(self, directory: Path | str):
        directory = Path(directory)
        paths = self.conf['projects'][self.project]['paths']

        for path_name, str_path in paths.items():
            source = directory / path_name
            dest = Path(str_path)
            copy_args = (str(source), str_path)
            if source.exists():
                if dest.is_file():
                    dest.unlink()
                elif dest.is_dir():
                    shutil.rmtree(str(dest))

                if source.is_file():
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(*copy_args)
                else:
                    shutil.copytree(*copy_args)

    def _get_commit_message(self) -> str:
        now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M')
        return f'Auto push from {now} [{self.project}]'



import functools
import re
import subprocess
import typing as t
from pathlib import Path

from config_keeper import config
from config_keeper import exceptions as exc
from config_keeper.console_helpers import (
    print_critical,
    print_error,
    print_warning,
)

TYPENAME: dict[type, str] = {
    str: 'string',
    dict: 'map',
}


path_name_regex = re.compile(r'^[\w-]+$')
path_regex = re.compile(r'^[\w\/~\-\. ]+$')


def ping_remote(
    repository: str,
    *,
    log: bool = False,
) -> subprocess.CompletedProcess[bytes] | subprocess.CompletedProcess[str]:
    cmd = ['git', 'ls-remote', repository, 'HEAD']
    if log:
        return subprocess.run(cmd, check=True, capture_output=True, text=True)
    return subprocess.run(
        cmd,
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )



def check_if_project_exists(project: str, conf: config.TConfig):
    if project not in conf['projects']:
        raise exc.ProjectDoesNotExistError(project)


def get_type(typehint: type | t.GenericAlias) -> type:
    return getattr(typehint, '__origin__', typehint)


ReportLevel = t.Literal['skip', 'warning', 'error', 'critical']


class Validator:
    def __init__(
        self,
        conf: config.TConfig,
    ):
        self.is_valid = True
        self.has_warnings = False
        self.conf = conf
        self._message_printers: list[t.Callable[[], None]] = []

    def print_errors(self):
        for printer in self._message_printers:
            printer()

    def _skip(self, _: str):
        pass

    def _warning(self, msg: str):
        self.has_warnings = True
        self._message_printers.append(functools.partial(print_warning, msg))

    def _error(self, msg: str):
        self.is_valid = False
        self._message_printers.append(functools.partial(print_error, msg))

    def _critical(self, msg: str):
        self.is_valid = False
        self._message_printers.append(functools.partial(print_critical, msg))

    def _report(self, report_type: str, msg: str):
        level: ReportLevel = getattr(self, report_type)
        getattr(self, f'_{level}')(msg)


RootReportType = t.Literal['unknown_param', 'type_mismatch']


class RootValidator(Validator):
    def __init__(
        self,
        conf: config.TConfig,
        *,
        unknown_param: ReportLevel = 'warning',
        type_mismatch: ReportLevel = 'critical',
    ):
        super().__init__(conf)
        self.unknown_param = unknown_param
        self.type_mismatch = type_mismatch

    def validate(self) -> bool:
        for param, value in self.conf.items():
            if typehint := config.TConfig.__annotations__.get(param, None):
                realtype = get_type(typehint)
                if not isinstance(value, realtype):
                    self._report('type_mismatch', (
                        f'"{param}" is not a {TYPENAME[realtype]}.'
                    ))
            else:
                self._report('unknown_param', f'unknown parameter "{param}".')

        return self.is_valid

    def _report(self, report_type: RootReportType, msg: str):
        super()._report(report_type, msg)


ProjectReportType = t.Literal[
    'path_existence',
    'repo_availability',
    'unknown_param',
    'missing_param',
    'empty_repository',
    'empty_branch',
    'empty_paths',
    'type_mismatch',
    'value_constraint',
]


class ProjectValidator(Validator):
    def __init__(
        self,
        conf: config.TConfig,
        *,
        path_existence: ReportLevel = 'error',
        repo_availability: ReportLevel = 'error',
        unknown_param: ReportLevel = 'warning',
        missing_param: ReportLevel = 'error',
        empty_repository: ReportLevel = 'error',
        empty_branch: ReportLevel = 'error',
        empty_paths: ReportLevel = 'warning',
        type_mismatch: ReportLevel = 'critical',
        value_constraint: ReportLevel = 'error',
    ):
        super().__init__(conf)
        self.path_existence = path_existence
        self.repo_availability = repo_availability
        self.unknown_param = unknown_param
        self.missing_param = missing_param
        self.empty_repository = empty_repository
        self.empty_branch = empty_branch
        self.empty_paths = empty_paths
        self.type_mismatch = type_mismatch
        self.value_constraint = value_constraint

    def validate(self, project: str) -> bool:
        check_if_project_exists(project, self.conf)

        if not isinstance(self.conf['projects'][project], dict):
            self._critical(
                f'"projects.{project}" is not a {TYPENAME[dict]}.',
            )
            return self.is_valid

        for param, value in sorted(self.conf['projects'][project].items()):
            self._validate_param(param, value, project)

        self._check_missing_params(project)

        return self.is_valid

    def _check_missing_params(self, project: str):
        required_params = set(config.TProject.__annotations__.keys())
        actual_params = set(self.conf['projects'][project].keys())
        for param in sorted(required_params - actual_params):
            self._report('missing_param', (
                f'"projects.{project}" missing parameter "{param}".'
            ))

    def _validate_param(
        self,
        param: str,
        value: t.Any,
        project: str,
    ):
        typehint = config.TProject.__annotations__.get(param, None)
        if not typehint:
            self._report('unknown_param', (
                f'unknown parameter "projects.{project}.{param}".'
            ))
            return

        realtype = get_type(typehint)
        if not isinstance(value, realtype):
            self._report('type_mismatch', (
                f'"projects.{project}.{param}" is not a {TYPENAME[realtype]}.'
            ))
            return

        if not value:
            self._report(f'empty_{param}', (
                f'"projects.{project}.{param}" is empty.'
            ))
            return

        if param == 'repository':
            self._validate_repository(value, project)
        elif param == 'paths':
            for path_name, path in sorted(value.items()):
                self._validate_path(path_name, path, project)

    def _validate_path(self, path_name: str, path: str, project: str):
        if not isinstance(path, str):
            self._report('type_mismatch', (
                f'"projects.{project}.paths.{path_name}" ({path}) is not a '
                f'{TYPENAME[str]}.'
            ))
        elif not path_name_regex.match(path_name):
            self._report('value_constraint', (
                f'"projects.{project}.paths.{path_name}" is not a valid path '
                'name.'
            ))
        elif not Path(path).exists():
            self._report('path_existence', (
                f'"projects.{project}.paths.{path_name}" ({path}) does not '
                'exist.'
            ))

    def _validate_repository(self, repository: t.Any, project: str):
        try:
            ping_remote(repository)
        except subprocess.CalledProcessError:
            self._report('repo_availability', (
                f'"projects.{project}.repository" ({repository}) is '
                'unavailable.'
            ))

    def _report(self, report_type: ProjectReportType, msg: str):
        super()._report(report_type, msg)



import typer

from config_keeper import config, console, settings
from config_keeper import exceptions as exc
from config_keeper.progress import spinner
from config_keeper.validation import ProjectValidator, RootValidator

cli = typer.Typer()


@cli.command()
def path():
    """
    Show configuration file path. You can set CONFIG_KEEPER_CONFIG_FILE
    environment variable to change its path.
    """

    config.ensure_exists()
    console.print(settings.CONFIG_FILE)


@cli.command()
def validate():
    """
    Validate config for missing or unknown params, check repositories and paths.
    """

    conf = config.load()
    is_valid = True
    has_warnings = False

    with spinner() as p:
        p.add_task('Validating...', total=None)

        root_validator = RootValidator(conf)
        is_valid = root_validator.validate()
        has_warnings = root_validator.has_warnings

        project_validator = ProjectValidator(conf, path_existence='warning')
        for project in conf['projects']:
            project_validator.validate(project)
            is_valid = is_valid and project_validator.is_valid
            has_warnings = has_warnings or project_validator.has_warnings

    root_validator.print_errors()
    project_validator.print_errors()

    if not is_valid:
        raise exc.InvalidConfigError

    if not has_warnings:
        console.print('OK', style='green')



import typing as t

import yaml

from config_keeper import settings


class TProject(t.TypedDict):
    repository: str
    branch: str
    paths: dict[str, str]


class TConfig(t.TypedDict):
    projects: dict[str, TProject]


def defaults() -> TConfig:
    return {
        'projects': {},
    }


def ensure_exists():
    settings.CONFIG_FILE.parent.mkdir(parents=True, exist_ok=True)
    settings.CONFIG_FILE.touch()


def load() -> TConfig:
    ensure_exists()
    from config_keeper import exceptions as exc
    try:
        config = yaml.load(
            settings.CONFIG_FILE.read_bytes(),
            Loader=yaml.Loader,
        )
    except yaml.error.YAMLError as e:
        msg = (
            f'{settings.CONFIG_FILE} is not a valid YAML file.\n'
            'Please fix or remove it.'
        )
        tip = (
            f'you can use\n> {settings.EXECUTABLE_NAME} config validate\n'
            'after.'
        )
        raise exc.InvalidConfigError(msg, tip=tip) from e
    if config is None:
        return defaults()
    if not isinstance(config, dict):
        msg = (
            f'the root object of {settings.CONFIG_FILE} config must be a map.\n'
            'Please fix or remove config.\n'
        )
        tip = (
            f'you can use\n> {settings.EXECUTABLE_NAME} config validate\n'
            'after.'
        )
        raise exc.InvalidConfigError(msg, tip=tip)
    return config


def save(config: TConfig):
    raw = yaml.dump(config)
    settings.CONFIG_FILE.write_text(raw)



import re
import typing as t
from pathlib import Path

import typer

from config_keeper import config
from config_keeper import exceptions as exc
from config_keeper.commands import helps
from config_keeper.console_helpers import print_project_saved
from config_keeper.validation import check_if_project_exists

cli = typer.Typer()


path_arg_regex = re.compile(r'(^[\w\-\. ]+):([\w\/~\-\. ]+$)')

overwrite_help = """
    If set, overwrite path names if project already has them. Fail otherwise.
"""
paths_help = """
    Path list with following syntax: path_name:/path/to/file/or/dir. Path names
    of each project must be unique as it will be used as temporary file
    (directory) name for storing it in repository. Path after colon is any
    path to file (directory) in your filesystem.
"""
path_names_help = """
    Path names of project.
"""
ignore_missing_help = """
    If set, path names that are not exist in project will be ignored. Fail
    otherwise.
"""


@cli.command()
def add(
    project: t.Annotated[str, typer.Option(help=helps.project)],
    paths: t.Annotated[
        t.List[str],  # noqa: UP006
        typer.Argument(help=paths_help),
    ],
    overwrite: t.Annotated[bool, typer.Option(help=overwrite_help)] = False,
):
    """
    Add directories or files to project. They will be pushed to (and pulled
    from) repository.
    """

    conf = config.load()
    check_if_project_exists(project, conf)

    new_paths = {}

    for raw_path in paths:
        match_ = path_arg_regex.match(raw_path)
        if not match_:
            msg = (
                f'{raw_path} is an invalid argument. Format must be as '
                'follows:\n\n'
                'key:/path/to/file/or/folder'
            )
            raise exc.InvalidArgumentFormatError(msg)
        key = match_.group(1)
        path = Path(match_.group(2))
        if key in new_paths:
            raise exc.DuplicatePathNameError(key)
        if key in conf['projects'][project]['paths'] and not overwrite:
            raise exc.PathNameAlreadyInProjectError(
                key,
                project,
                tip='you can use --overwrite option.',
            )
        new_paths[key] = str(path.expanduser().resolve())

    conf['projects'][project]['paths'].update(new_paths)
    config.save(conf)
    print_project_saved(project)


@cli.command()
def delete(
    project: t.Annotated[str, typer.Option(help=helps.project)],
    path_names: t.Annotated[
        t.List[str],   # noqa: UP006
        typer.Argument(help=path_names_help),
    ],
    ignore_missing: t.Annotated[
        bool,
        typer.Option(help=ignore_missing_help),
    ] = False,
):
    """
    Delete paths by their path names from project. This will not affect original
    files or directories.
    """

    conf = config.load()
    check_if_project_exists(project, conf)

    for name in path_names:
        try:
            del conf['projects'][project]['paths'][name]
        except KeyError as e:
            if ignore_missing:
                continue
            raise exc.PathNameDoesNotExistError(
                name,
                project,
                tip=(
                    'you can use --ignore-missing option to suppress these '
                    'errors.'
                ),
            ) from e

    config.save(conf)
    print_project_saved(project)



import typing as t
from importlib import import_module
from pathlib import Path


class LazySettings:
    CONFIG_FILE: Path
    EXECUTABLE_NAME: str

    def __init__(self, settings_module: str):
        _super = super()
        _super.__setattr__('settings_module', import_module(settings_module))
        _super.__setattr__('real_settings', {
            key: getattr(self.settings_module, key)
            for key in dir(self.settings_module)
            if key.isupper()
        })
        _super.__setattr__('overridden_settings', {})

    def __getattr__(self, key: str) -> t.Any:
        return self.overridden_settings.get(key, self.real_settings[key])

    def __setattr__(self, key: str, value: t.Any):
        if key not in self.real_settings:  # nocv
            msg = f'{self.settings_module} has no attribute {key}'
            raise AttributeError(msg)
        self.overridden_settings[key] = value



import os
from pathlib import Path

# paths
_DEFAULT_CONFIG_FILE = Path.home() / '.config-keeper.yaml'
CONFIG_FILE = Path(os.getenv('CONFIG_KEEPER_CONFIG_FILE', _DEFAULT_CONFIG_FILE))

# etc
EXECUTABLE_NAME = 'config-keeper'
